﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Programmable {
  [StorableClass]
  public abstract class ProgrammableProblem<TEncoding, TEvaluator> : HeuristicOptimizationProblem<TEvaluator, ISolutionCreator>, IProblemDefinition, IStorableContent
    where TEncoding : class, IEncoding
    where TEvaluator : class, IEvaluator {

    public string Filename { get; set; }

    protected IValueParameter<TEncoding> EncodingParameter {
      get { return (IValueParameter<TEncoding>)Parameters["Encoding"]; }
    }

    IEncoding IProblemDefinition.Encoding { get { return Encoding; } }
    public TEncoding Encoding {
      get { return EncodingParameter.Value; }
      protected set {
        if (value == null) throw new ArgumentNullException("Encoding must not be null.");
        if (Encoding != null) Parameters.RemoveRange(Encoding.Parameters);
        EncodingParameter.Value = value;
        Parameters.AddRange(Encoding.Parameters);
      }
    }

    public virtual IEnumerable<Individual> GetNeighbors(Individual individual, IRandom random) {
      return Enumerable.Empty<Individual>();
    }

    protected override IEnumerable<IItem> GetOperators() {
      return base.GetOperators().Concat(Encoding.Operators);
    }
    public override IEnumerable<IParameterizedItem> ExecutionContextItems {
      get { return base.ExecutionContextItems.Concat(new[] { Encoding }); }
    }

    public ProgrammableProblem()
      : base() {
      Parameters.Add(new ValueParameter<TEncoding>("Encoding", "Describes the configuration of the encoding, what the variables are called, what type they are and their bounds if any."));
      RegisterEvents();
    }

    protected ProgrammableProblem(ProgrammableProblem<TEncoding, TEvaluator> original, Cloner cloner)
      : base(original, cloner) {
      RegisterEvents();
    }

    [StorableConstructor]
    protected ProgrammableProblem(bool deserializing)
      : base(deserializing) {
      RegisterEvents();
    }

    private void RegisterEvents() {
      EncodingParameter.ValueChanged += (o, e) => OnEncodingChanged();
    }

    protected virtual void OnEncodingChanged() {
      var solutionCreatorType = Encoding.GetType().BaseType.GetGenericArguments();
      var paramType = typeof(ValueParameter<>).MakeGenericType(solutionCreatorType);
      var solutionCreatorParam = (IParameter)Activator.CreateInstance(paramType, SolutionCreatorParameter.Name, SolutionCreatorParameter.Description,
        Encoding.SolutionCreator);
      Parameters.Remove(SolutionCreatorParameter);
      Parameters.Add(solutionCreatorParam);
    }

    protected override void OnSolutionCreatorChanged() {
      base.OnSolutionCreatorChanged();
      Encoding.SolutionCreator = SolutionCreator;
    }


    protected virtual void ParameterizeOperators() {
      foreach (var op in Operators.OfType<IEncodingOperator>())
        op.EncodingParameter.ActualName = EncodingParameter.Name;
      foreach (var op in Operators.OfType<INeighborBasedOperator>())
        op.GetNeighborsFunc = GetNeighbors;
    }

  }
}
